---------------------------------------------------------------------------------------------
-- QHRaid Class
---------------------------------------------------------------------------------------------
QHRaid = {};
QHRaid.__index = QHRaid;

function QHRaid:new(id)

	-- Metatable
	local self = {};
     	setmetatable(self, QHRaid);
     	
     	-- Position Variables
	self.x = 0;
	self.y = 0;
	self.align = "TOPLEFT";
	self.relative = "TOPLEFT";
	self.anchor = "UIParent";
	self.width = 0;
	self.height = 0;
	self.moving = nil;
	
	-- Frame Variables
	self.id = id;
	self.name = QH:Concat("QHRaid", id);
	self.globalframe = CreateFrame("Frame", self.name, UIParent, "QHRAIDTEMPLATE");
	self.frame = getglobal(QH:Concat(self.globalframe:GetName(), "_Frame"));

	-- Background Variables
	self.background = getglobal(QH:Concat(self.frame:GetName(), "Background"));
	self.backdrop = {bgFile = "", edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
			 tile = true, tileSize = 4, edgeSize = 8,
			 insets = { left = 2, right = 2, top = 2, bottom = 2 }};
	
	-- Raid Unit Variables
	self.raidunit = {};
	self.notinraid = {};
	self.raidsorted = {};
	self.forceupdate = true;
	for n = 1,40 do
		self.raidunit[n] = QHUnit:new(QH:Concat("raid", n));
		self.raidunit[n].parent = self;
		self.raidunit[n]:SetPosition(0, 0, "CENTER", "CENTER", self.frame);
		self.raidsorted[n] = self.raidunit[n];
	end
	
	-- Reference Variables
	self.obj = self;
	self.parent = nil;
	self.frame.obj = self;
	self.globalframe.obj = self;
	
	-- Call Update Function and Return
	self:Update();
	return self;
end

---------------------------------------------------------------------------------------------
-- QHRaid Methods
---------------------------------------------------------------------------------------------
function QHRaid:SetPosition(x, y, align, relative, anchor)

	-- Set Position
	if x then self.x = x; end
	if y then self.y = y; end
	if align then self.align = align; end
	if relative then self.relative = relative; end
	if anchor then 
		if type(anchor) == string then self.anchor = anchor;
		elseif anchor:GetName() then self.anchor = anchor:GetName(); end
	end
	
	-- Save Position
	QH:SetSetting("Raid_Position_X", x);
	QH:SetSetting("Raid_Position_Y", y);
	QH:SetSetting("Raid_Position_Align", align);
	QH:SetSetting("Raid_Position_Anchor", anchor);
	QH:SetSetting("Raid_Position_Relative", relative);
end

function QHRaid:ReSize()

	local pad_width = 8;
	local pad_height = 4;
	local maxrows = QH:GetSetting("Raid_Rows");

	local w = 0;
	local h = 0;
	for row,value in pairs(self.raidsorted) do
	
		local col = floor((row - 1) / maxrows);
		if value and value.unitname then
		
			if row <= maxrows then
				h = h + value.height + QH:GetSetting("Raid_Row_Spacing");
			end
			w = max(((col + 1) * value.width) + (QH:GetSetting("Raid_Col_Spacing") * col), w);
		end
	end
	
	self.width = w + pad_width + QH:GetSetting("Raid_Col_Offset");
	self.height = h + pad_height + QH:GetSetting("Raid_Row_Offset");
	return w, h;
end

function QHRaid:Sort()

	local method = QH:GetSetting("Raid_Sort_By");
	local groups = QH:GetSetting("Raid_Groups");
	local dynamic_groups = QH:GetSetting("Raid_Dynamic_Groups");
	if dynamic_groups then
		local _, _, _, _, instance_max = GetInstanceInfo();
		if instance_max > 0 then groups = "1;"; 
		else groups = "1;2;3;4;5;6;7;8;"; end
		if instance_max > 5 then groups = QH:Concat(groups, "2;"); end
		if instance_max > 10 then groups = QH:Concat(groups, "3;"); end
		if instance_max > 15 then groups = QH:Concat(groups, "4;"); end
		if instance_max > 20 then groups = QH:Concat(groups, "5;"); end
		if instance_max > 25 then groups = QH:Concat(groups, "6;"); end
		if instance_max > 30 then groups = QH:Concat(groups, "7;"); end
		if instance_max > 35 then groups = QH:Concat(groups, "8;"); end
	end	
	
	-- Clean Table
	table.wipe(self.notinraid);
	table.wipe(self.raidsorted);
	for n = 1,40 do
		local unit = self.raidunit[n];
		local name, rank, subgroup = GetRaidRosterInfo(tonumber(n));
		if unit then
			if UnitExists(unit.unitid) and string.find(groups, subgroup) then
				table.insert(self.raidsorted, unit);
			else
				table.insert(self.notinraid, unit);
			end
		end
	end
	
	-- Sort Units
	if(method == "CLASS") then
	
		function CompareClass(a, b)
		
			if not a or not a.unitclass then return false; end
			if not b or not b.unitclass then return true; end
			
			local order = QH:GetSetting("Raid_Class_Order");
			local ai = string.find(order, a.unitclass);
			local bi = string.find(order, b.unitclass);
			
			if not ai then return false; end
			if not bi then return true; end
			
			if ai == bi then 
				if not a.unitname then return false; end
				if not b.unitname then return true; end
				return a.unitname < b.unitname; 
			end
		
			return ai < bi;
		end
		table.sort(self.raidsorted, CompareClass);
		
		-- Add Class Labels/Breaks
		
	elseif(method == "GROUP") then
	
		function CompareGroup(a, b)
			
			if not a or not a.unitid then return false; end
			if not b or not b.unitid then return true; end
			
			local ai = string.gsub(a.unitid, "raid", "");
			local bi = string.gsub(b.unitid, "raid", "");
			local namea, ranka, subgroupa = GetRaidRosterInfo(tonumber(ai));
			local nameb, rankb, subgroupb = GetRaidRosterInfo(tonumber(bi));
			
			if subgroupa == subgroupb then 
				if not a.unitname then return false; end
				if not b.unitname then return true; end
				return a.unitname < b.unitname;
			end
			return subgroupa < subgroupb;
		end
		table.sort(self.raidsorted, CompareGroup);
		
		-- Add Group Labels/Breaks
	end
end

function QHRaid:GetUnitByName(name)
	for index, unit in pairs(self.raidunit) do
		if UnitName(unit) == name then return unit.unitid; end
	end
end

function QHRaid:StartMoving(button)

	if self.obj.parent then
		self.obj.parent:StartMoving(button);
	elseif IsShiftKeyDown() then
		self.obj.moving = true;
		self.obj.frame:SetMovable(true);
		self.obj.frame:StartMoving();
	end
end

function QHRaid:StopMoving()

	if self.obj.parent then
		self.obj.parent:StopMoving();
	else
		self.obj.frame:StopMovingOrSizing();
		align, anchor, relative, x, y = self.frame:GetPoint();
		if not anchor then anchor = UIParent; end
		self.obj:SetPosition(x, y, align, relative, anchor);
		self.obj.moving = nil;
	end
end

function QHRaid:Hide()

	if self.frame:IsVisible() then self.frame:Hide(); end
end

function QHRaid:Show()
	
	if not self.frame:IsVisible() then
	
		-- Show Raid Frame
		self.frame:Show();
		self.forceupdate = true;
	end
end

---------------------------------------------------------------------------------------------
-- QHRaid Events
---------------------------------------------------------------------------------------------
function QHRaid:ForceUpdate()
	self.forceupdate = true;
	self:Update();
end

function QHRaid:Update()

	if (UnitInRaid("player") or GetNumRaidMembers() > 0) and QH:GetSetting("Raid_Frame_Show") then

		-- Get Settings
		local hide_party = QH:GetSetting("Raid_Party_Hide");
		local background_color = QH:GetSetting("Raid_Background_Color");
		local background_show = QH:GetSetting("Raid_Background_Show");
		self.align = QH:GetSetting("Raid_Position_Align");
		self.anchor = getglobal(QH:GetSetting("Raid_Position_Anchor"));
		self.relative = QH:GetSetting("Raid_Position_Relative");
		self.x = QH:GetSetting("Raid_Position_X");
		self.y = QH:GetSetting("Raid_Position_Y");

		-- Secure Code
		if not InCombatLockdown() then

			-- Show Raid Frame
			self:Show();

			-- Hide Party Frame In Raid
			local oldpartyinterface = HIDE_PARTY_INTERFACE;
			if hide_party then

				-- Hide Party Bars
				getglobal("QHUnit_party1"):Hide();
				getglobal("QHUnit_party2"):Hide();
				getglobal("QHUnit_party3"):Hide();
				getglobal("QHUnit_party4"):Hide();
				setglobal("HIDE_PARTY_INTERFACE", "1");
			else
				setglobal("HIDE_PARTY_INTERFACE", "0");
			end
			if(oldpartyinterface ~= HIDE_PARTY_INTERFACE) then RaidOptionsFrame_UpdatePartyFrames(); end
			
			-- Force Unit Position Updates
			if self.forceupdate then
				self:UpdatePositions();
			end

			-- Set Frame Position
			if not self.moving and self.align and self.x and self.y then
				self.frame:ClearAllPoints();
				self.frame:SetPoint(self.align, self.anchor, self.relative, self.x, self.y);
			end

			-- Set Frame Background
			if background_show then
				self.frame:SetBackdrop(self.backdrop);
				self.background:SetTexture(background_color.r, background_color.g, background_color.b, background_color.a); 
			else
				self.frame:SetBackdrop(nil);
				self.background:SetTexture(0, 0, 0, 0); 
			end

			-- Set Background Frame Size
			self:ReSize();
			if self.width and self.height then
				self.frame:SetWidth(self.width);
				self.frame:SetHeight(self.height);
			end
		end
		
		-- Update Range and Raid Icons
		for index, value in pairs(self.raidsorted) do
			value:UpdateRange();
			value:UpdateRaidIcon();
		end
		
		-- Reset Old Unit Frames
		for index, value in pairs(self.notinraid) do
			value:Clear();
		end

	elseif self.frame:IsVisible() then

		-- Clear Unit Frames
		for index,value in pairs(self.raidunit) do
			value:Clear();
			value:Hide();
		end

		-- Hide Raid Frame
		if not InCombatLockdown() then
			self:Hide();
		end
	end
end

function QHRaid:UpdatePositions()

	if UnitInRaid("player") or GetNumRaidMembers() > 0 then

		-- Get Settings
		local hide_party = QH:GetSetting("Raid_Party_Hide");
		local background_color = QH:GetSetting("Raid_Background_Color");
		local background_show = QH:GetSetting("Raid_Background_Show");
		self.align = QH:GetSetting("Raid_Position_Align");
		self.anchor = getglobal(QH:GetSetting("Raid_Position_Anchor"));
		self.relative = QH:GetSetting("Raid_Position_Relative");
		self.x = QH:GetSetting("Raid_Position_X");
		self.y = QH:GetSetting("Raid_Position_Y");

		-- Secure Code
		if not InCombatLockdown() then
		
			-- Sort Raid
			self:Sort();
			self.forceupdate = false;
			
			-- Update Unit Positions
			for index, value in pairs(self.raidsorted) do
				if value then
				
					local row = index;
					local rows = QH:GetSetting("Raid_Rows"); --26
					local rowspacing = QH:GetSetting("Raid_Row_Spacing"); --3
					local rowoffset = QH:GetSetting("Raid_Row_Offset"); --8
					local colspacing = QH:GetSetting("Raid_Col_Spacing"); --25
					local coloffset = QH:GetSetting("Raid_Col_Offset"); --8
					
					-- Reset Unit Info
					value:LoadUnitInfo();
					value:Resize();

					local y = -((mod(row-1,rows))*(value.height + rowspacing)) - rowoffset;
					local x = (floor((row-1) / rows) * (value.width + colspacing)) + coloffset;
					
					-- Update Unit Frame
					value:SetPosition(x, y, "TOPLEFT", "TOPLEFT", self.frame);
					value:Show();
					value:Update();
				end
			end
			
			-- Reset Old Unit Frames
			for index, value in pairs(self.notinraid) do
				value:SetPosition(-100, -100, "TOPLEFT", "TOPLEFT", self.frame);
				value:Hide();
				value:Update();
			end
		else
			-- Sort When Out of Combat
			self.forceupdate = true;
		end
	end
end
